#pragma once

// taskflow.hpp
// ^
// |
// syclflow.hpp

#include "taskflow.hpp"
#include "sycl/sycl_flow.hpp"
#include "sycl/sycl_algorithm/sycl_for_each.hpp"
#include "sycl/sycl_algorithm/sycl_reduce.hpp"
#include "sycl/sycl_algorithm/sycl_transform.hpp"

/**
@file syclflow.hpp
@brief main syclFlow include file
*/
